{
  "type": "record",
  "name": "DatasetKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a Dataset",
  "fields": [
    {
      "name": "platform",
      "type": "string",
      "doc": "Data platform urn associated with the dataset",
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "name",
      "type": "string",
      "doc": "Dataset native name e.g. <db>.<table>, /dir/subdir/<name>, or <name>",
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "origin",
      "type": {
        "type": "enum",
        "name": "FabricType",
        "namespace": "com.linkedin.pegasus2avro.common",
        "doc": "Fabric group type",
        "symbols": [
          "DEV",
          "EI",
          "PROD",
          "CORP"
        ],
        "symbolDocs": {
          "CORP": "Designates corporation fabrics",
          "DEV": "Designates development fabrics",
          "EI": "Designates early-integration (staging) fabrics",
          "PROD": "Designates production fabrics"
        }
      },
      "doc": "Fabric type where dataset belongs to or where it was generated.",
      "Searchable": {
        "addToFilters": true,
        "fieldType": "TEXT_PARTIAL",
        "filterNameOverride": "Environment",
        "queryByDefault": false
      }
    }
  ],
  "Aspect": {
    "name": "datasetKey"
  }
}